<?php
/**
 * A class to manage cascadable paths through arrays
 *
 * Revision summary:
 *
 * 2011-10-26	Initial version by Brian Katzung
 *
 * @copyright Copyright 2011 Brian Katzung and Kappa Computer Solutions, LLC
 * @author Brian Katzung <briank@kappacs.com>
 * @license http://www.opensource.org/licenses/MIT MIT License
 * @package Array
 */

class ArrayPath {

    /**
     * @var boolean Absolute (true) or relative (false) path.
     *
     * (Absolute paths do not prepend their viewpoints during resolution.)
     */
    protected $isAbsolute;

    /**
     * @var boolean|null Is this an append path (true; contains '[]') or
     *  a specific path (false)?
     */
    protected $isAppend;

    /**
     * @var object|null This path's viewpoint (cascade prefix).
     *
     * The viewpoint's path is prepended for relative paths
     * (like a current directory while browsing files).
     */
    protected $viewpoint;

    /**
     * @var array The parts (components) of the path.
     */
    protected $parts;

    /**
     * Class methods
     */

    /**
     * Constructor
     *
     * @param string|array $path A (partial) path through an array.
     * @param boolean $isAbsolute Is this an absolute (true) or a relative
     *  (false; the default) path?
     */
    public function __construct ($path, $isAbsolute = false) {
	$this->isAbsolute = $isAbsolute;
	$this->viewpoint = null;
	if (is_array($path)) { $this->loadParts($path); }
	else { $this->parse($path); }
    }

    /**
     * A convenience function to create and initialize an ArrayPath.
     *
     * If passed an existing path object, it will be retained. If
     * passed a string or array path, it will used to generate a new
     * ArrayPath.
     *
     * If a viewpoint is provided, the viewpoint will be set unless an
     * existing object already has a viewpoint.
     *
     * @param mixed $path The array path object, string, or array.
     * @param mixed|null $viewpoint An optional viewpoint.
     * @return ArrayPath|object
     */
    static public function get ($path, $viewpoint = null, $isAbsolute = false) {
	if (!is_object($path)) {
	    $class = __CLASS__;
	    $path = new $class($path, $isAbsolute);
	}

	if (!$path->getViewpoint() && isset($viewpoint)) {
	    $path->setViewpoint(self::get($viewpoint));
	}

	return $path;
    }

    /**
     * Instance methods
     */

    /**
     * Is this an absolute path?
     *
     * @return boolean
     */
    public function isAbsolute () { return $this->isAbsolute; }

    /**
     * Is this an append path?
     *
     * @param null|boolean $append Optionally change this setting.
     * @return boolean
     */
    public function isAppend ($append = null) {
	if (isset($append)) { $this->isAppend = !!$append; }
	return $this->isAppend;
    }

    /**
     * Load an array into path components.
     *
     * @param array $path The path components.
     * @return ArrayPath
     */
    protected function loadParts ($path) {
	$parts = array();
	$this->isAppend = false;
	foreach ($path as $part) {
	    if ('[]' === $part) {
		$this->isAppend = true;
	    } else {
		$parts[] = $part;
	    }
	}
	$this->parts = $parts;
	return $this;
    }

    /**
     * Parse a string into path components.
     *
     * @param string $path The path string.
     * @return ArrayPath
     */
    protected function parse ($path) {
	preg_match_all('/[^\[\]]+|\[\]/', $path, $parts);
	return $this->loadParts($parts[0]);
    }

    /**
     * Return some or all of the path components.
     *
     * @param boolean $all Include append parts ('[]') if true; the
     *  default is false.
     * @return array
     */
    public function parts ($all = false) {
	if ($this->getViewpoint() && !$this->isAbsolute()) {
	    $out = array_merge($this->getViewpoint()->parts(), $this->parts);
	} else {
	    $out = $this->parts;
	}

	if ($all && $this->isAppend) { $out[] = '[]'; }
	return $out;
    }

    /**
     * Return the resolved path as a string.
     *
     * @param boolean $all Include append parts ('[]') if true
     *  (the default).
     * @return string
     */
    public function toString ($all = true) {
	$parts = array();
	foreach ($this->parts($all) as $part) {
	    if (!isset($parts[0]) || '[]' === $part) {
		$parts[] = $part;
	    } else {
		$parts[] = "[$part]";
	    }
	}
	return join('', $parts);
    }

    /**
     * Traverse an array, applying the resolved path.
     *
     * @param array &$base A reference to the array (base).
     * @param array|null $options An optional array of options:
     *
     * create - Create missing arrays during traversal if true <br>
     * partial - Don't traverse the last path component if true <br>
     * last - Where to return the last part by reference on a partial <br>
     * needArray - Throw an exception when resolving to a scalar if true
     *
     * @return array
     */
    public function &apply (&$base, $options = null) {
	$options = (array) $options;
	$parts = $this->parts();
	if ($options['partial']) {
	    $options['last'] = array_pop($parts);
	}

	$current = &$base;
	foreach ($parts as $part) {
	    if (!is_array($current)) {
		throw new Exception("Cannot use scalar value as an array");
	    }
	    if (isset($current[$part])) {
		$current = &$current[$part];
	    } else if ($options['create']) {
		$current[$part] = array();
		$current = &$current[$part];
	    } else {
		unset($current);
		return $current;
	    }
	}

	if ($options['needArray'] && !is_array($current)) {
	    throw new Exception("Array path did not resolve to an array");
	}
	return $current;
    }

    /**
     * Get this path's current viewpoint.
     *
     * @return ArrayPath
     */
    public function getViewpoint () { return $this->viewpoint; }

    /**
     * Set this path's current viewpoint.
     *
     * @param object|array|string $viewpoint The new viewpoint.
     * @return ArrayPath
     */
    public function setViewpoint ($viewpoint) {
	$this->viewpoint = self::get($viewpoint);
	return $this;
    }

}

# END
